@@ -40,11 +40,17 @@
  */
 #define _XOPEN_SOURCE 700
 
+#if defined(HAVE_ISSETUGID) && defined(__FreeBSD__)
+#define __BSD_VISIBLE 1
+#endif
+
 #include "compat.h"
+#include "debug.h"
 
 #include <assert.h>
 #include <dirent.h>
 #include <errno.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -163,18 +169,20 @@ getprogname (void)
 #include <fcntl.h>
 #include <unistd.h>
 
+#ifndef PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
 void
-p11_mutex_init (p11_mutex_t *mutex)
+p11_recursive_mutex_init (p11_mutex_t *mutex)
 {
 	pthread_mutexattr_t attr;
 	int ret;
 
 	pthread_mutexattr_init (&attr);
-	pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_DEFAULT);
+	pthread_mutexattr_settype (&attr, PTHREAD_MUTEX_RECURSIVE);
 	ret = pthread_mutex_init (mutex, &attr);
 	assert (ret == 0);
 	pthread_mutexattr_destroy (&attr);
 }
+#endif
 
 char *
 p11_dl_error (void)
@@ -273,12 +281,6 @@ p11_dl_error (void)
 	return msg_buf;
 }
 
-void
-p11_dl_close (void *dl)
-{
-	FreeLibrary (dl);
-}
-
 int
 p11_thread_create (p11_thread_t *thread,
                    p11_thread_routine routine,
@@ -488,6 +490,23 @@ strndup (const char *data,
 
 #endif /* HAVE_STRNDUP */
 
+#ifndef HAVE_REALLOCARRAY
+
+void *
+reallocarray (void *ptr,
+	      size_t nmemb,
+	      size_t size)
+{
+	assert (nmemb > 0 && size > 0);
+	if (SIZE_MAX / nmemb < size) {
+		errno = ENOMEM;
+		return NULL;
+	}
+	return realloc (ptr, nmemb * size);
+}
+
+#endif /* HAVE_MEMDUP */
+
 #ifndef HAVE_STRCONCAT
 
 #include <stdarg.h>
@@ -503,8 +522,14 @@ strconcat (const char *first,
 
 	va_start (va, first);
 
-	for (arg = first; arg; arg = va_arg (va, const char*))
-	       length += strlen (arg);
+	for (arg = first; arg; arg = va_arg (va, const char*)) {
+		size_t old_length = length;
+		length += strlen (arg);
+		if (length < old_length) {
+			va_end (va);
+			return_val_if_reached (NULL);
+		}
+	}
 
 	va_end (va);
 
@@ -664,7 +689,7 @@ _gettemp (char *path,
 
 	/* Fill space with random characters */
 	while (trv >= path && *trv == 'X') {
-		rnd = rand () % sizeof (padchar) - 1;
+		rnd = rand () % (sizeof (padchar) - 1);
 		*trv-- = padchar[rnd];
 	}
 	start = trv + 1;
@@ -785,7 +810,7 @@ getauxval (unsigned long type)
 		extern int __libc_enable_secure;
 		secure = __libc_enable_secure;
 
-#elif defined(HAVE_ISSETUGID)
+#elif defined(HAVE_ISSETUGID) && defined(HAVE_ISSETUGID_OPENBSD)
 		secure = issetugid ();
 
 #elif defined(OS_UNIX)
@@ -853,6 +878,17 @@ strerror_r (int errnum,
 
 #endif /* HAVE_STRERROR_R */
 
+void
+p11_dl_close (void *dl)
+{
+#ifdef OS_WIN32
+	FreeLibrary (dl);
+#else
+	(void) dlclose (dl);
+#endif
+}
+
+
 #ifdef OS_UNIX
 
 #include <unistd.h>
@@ -879,6 +915,7 @@ fdwalk (int (* cb) (void *data, int fd),
 	struct rlimit rl;
 #endif
 
+#ifdef __linux__
 	dir = opendir ("/proc/self/fd");
 	if (dir != NULL) {
 		while ((de = readdir (dir)) != NULL) {
@@ -901,6 +938,7 @@ fdwalk (int (* cb) (void *data, int fd),
 		closedir (dir);
 		return res;
 	}
+#endif
 
 	/* No /proc, brute force */
 #ifdef HAVE_SYS_RESOURCE_H
@@ -922,3 +960,19 @@ fdwalk (int (* cb) (void *data, int fd),
 #endif /* HAVE_FDWALK */
 
 #endif /* OS_UNIX */
+
+int
+p11_ascii_tolower (int c)
+{
+	if (c >= 'A' && c <= 'Z')
+		return 'a' + (c - 'A');
+	return c;
+}
+
+int
+p11_ascii_toupper (int c)
+{
+	if (c >= 'a' && c <= 'z')
+		return 'A' + (c - 'a');
+	return c;
+}
