@@ -37,6 +37,7 @@
 #define P11_DEBUG_FLAG P11_DEBUG_TOOL
 
 #include "attrs.h"
+#include "constants.h"
 #include "debug.h"
 #include "oid.h"
 #include "dict.h"
@@ -268,29 +269,32 @@ extract_certificate (p11_enumerate *ex)
 	return true;
 }
 
+static CK_ATTRIBUTE *
+prepare_attr_types (void)
+{
+	CK_ATTRIBUTE *attrs;
+	int i, count;
+
+	/* Count the number of attributes we know about */
+	for (count = 0; p11_constant_types[count].value != CKA_INVALID; count++);
+
+	attrs = calloc (count + 1, sizeof (CK_ATTRIBUTE));
+	return_val_if_fail (attrs != NULL, NULL);
+
+	for (i = 0; i < count; i++)
+		attrs[i].type = p11_constant_types[i].value;
+	attrs[count].type = CKA_INVALID;
+
+	return attrs;
+}
+
 static bool
 extract_info (p11_enumerate *ex)
 {
 	CK_ATTRIBUTE *attr;
 	CK_RV rv;
 
-	static const CK_ATTRIBUTE attr_types[] = {
-		{ CKA_ID, },
-		{ CKA_CLASS, },
-		{ CKA_CERTIFICATE_TYPE, },
-		{ CKA_LABEL, },
-		{ CKA_VALUE, },
-		{ CKA_SUBJECT, },
-		{ CKA_ISSUER, },
-		{ CKA_SERIAL_NUMBER, },
-		{ CKA_TRUSTED, },
-		{ CKA_CERTIFICATE_CATEGORY },
-		{ CKA_X_DISTRUSTED },
-		{ CKA_PUBLIC_KEY_INFO },
-		{ CKA_INVALID, },
-	};
-
-	ex->attrs = p11_attrs_dup (attr_types);
+	ex->attrs = prepare_attr_types ();
 	rv = p11_kit_iter_load_attributes (ex->iter, ex->attrs, p11_attrs_count (ex->attrs));
 
 	/* The attributes couldn't be loaded */
@@ -304,24 +308,26 @@ extract_info (p11_enumerate *ex)
 		return false;
 
 	/* If a certificate then  */
-	if (ex->klass != CKO_CERTIFICATE) {
-		p11_message ("skipping non-certificate object");
-		return false;
-	}
+	if (ex->flags & P11_ENUMERATE_CORRELATE) {
+		if (ex->klass != CKO_CERTIFICATE) {
+			p11_message ("skipping non-certificate object");
+			return false;
+		}
 
-	if (!extract_certificate (ex))
-		return false;
+		if (!extract_certificate (ex))
+			return false;
 
-	attr = p11_attrs_find_valid (ex->attrs, CKA_PUBLIC_KEY_INFO);
-	if (attr) {
-		ex->attached = load_attached_extensions (ex, attr);
-		if (!ex->attached)
+		attr = p11_attrs_find_valid (ex->attrs, CKA_PUBLIC_KEY_INFO);
+		if (attr) {
+			ex->attached = load_attached_extensions (ex, attr);
+			if (!ex->attached)
+				return false;
+		}
+
+		if (!extract_purposes (ex))
 			return false;
 	}
 
-	if (!extract_purposes (ex))
-		return false;
-
 	return true;
 }
 
@@ -368,6 +374,11 @@ on_iterate_load_filter (p11_kit_iter *iter,
 	if (ex->limit_to_purposes && ex->purposes) {
 		*matches = CK_FALSE;
 		for (i = 0; i < ex->purposes->num; i++) {
+			if (strcmp (ex->purposes->elem[i], P11_OID_ANY_EXTENDED_KEY_USAGE_STR) == 0) {
+				p11_debug ("anyExtendedKeyUsage is set, skipping filtering by purposes");
+				*matches = CK_TRUE;
+				break;
+			}
 			if (p11_dict_get (ex->limit_to_purposes, ex->purposes->elem[i])) {
 				*matches = CK_TRUE;
 				break;
@@ -593,12 +604,16 @@ p11_enumerate_opt_filter (p11_enumerate *ex,
 		attrs = p11_attrs_build (NULL, &certificate, &x509, NULL);
 		ex->flags |= P11_ENUMERATE_COLLAPSE;
 
+	} else if (strcmp (option, "all") == 0) {
+		attrs = p11_attrs_build (NULL, NULL);
+
 	} else {
 		p11_message ("unsupported or unrecognized filter: %s", option);
 		return false;
 	}
 
 	p11_kit_iter_add_filter (ex->iter, attrs, p11_attrs_count (attrs));
+	p11_attrs_free (attrs);
 	ex->num_filters++;
 	return true;
 }
@@ -659,6 +674,8 @@ p11_enumerate_opt_purpose (p11_enumerate *ex,
 	return true;
 }
 
+extern bool p11_print_messages;
+
 bool
 p11_enumerate_ready (p11_enumerate *ex,
                      const char *def_filter)
@@ -672,8 +689,13 @@ p11_enumerate_ready (p11_enumerate *ex,
 	 * We only "believe" the CKA_TRUSTED and CKA_X_DISTRUSTED attributes
 	 * we get from modules explicitly marked as containing trust-policy.
 	 */
-	if (!ex->modules)
-		ex->modules = p11_kit_modules_load_and_initialize (P11_KIT_MODULE_TRUSTED);
+	if (!ex->modules) {
+		int flags = P11_KIT_MODULE_TRUSTED;
+		if (p11_print_messages)
+			flags |= P11_KIT_MODULE_VERBOSE;
+
+		ex->modules = p11_kit_modules_load_and_initialize (flags);
+	}
 	if (!ex->modules)
 		return false;
 	if (ex->modules[0] == NULL)
